from os.path import dirname
from sys import path as syspath
from typing import Iterable

from hatchling.builders.config import BuilderConfig
from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from packaging.tags import Tag

# During env initialization the PYTHONPATH doesn't include our helpers, so
# give it some help:
syspath.append(dirname(__file__))
from helpers.v8_build import build_v8, clean_v8, get_platform_tag  # noqa: E402


class V8BuildHook(BuildHookInterface[BuilderConfig]):
    def clean(self, versions: Iterable[str]) -> None:
        del versions

        clean_v8("src/py_mini_racer")

    def initialize(self, version: str, build_data):
        del version

        artifacts = build_v8(out_path="src/py_mini_racer")

        build_data.setdefault("force_include", {}).update(artifacts)

        # From https://stackoverflow.com/questions/76450587/python-wheel-that-includes-shared-library-is-built-as-pure-python-platform-indep
        # We have to tell Hatch that we're building a non-pure-Python wheel, even
        # though there are no extension modules (instead, there is native code in a
        # dynamic-link library in the package):
        build_data["pure_python"] = False

        # Because we aren't building an extension module (just pure Python and a
        # Python-independent DLL), any single wheel we create is broadly compatible
        # with different Python interpreters. Just mark py3 with any ABI:
        tag = Tag("py3", "none", get_platform_tag())

        build_data["tag"] = str(tag)



from argparse import ArgumentParser
from errno import EEXIST
from functools import lru_cache
from logging import DEBUG, basicConfig, getLogger
from os import environ, makedirs, pathsep, remove, symlink, unlink
from os.path import abspath, dirname, exists, isdir, isfile
from os.path import join as pathjoin
from platform import machine
from re import match
from shlex import join as shlexjoin
from shutil import copyfile, rmtree
from subprocess import check_call
from sys import executable, platform

from packaging.tags import platform_tags

basicConfig()
LOGGER = getLogger(__name__)
LOGGER.setLevel(DEBUG)
ROOT_DIR = dirname(abspath(__file__))
V8_VERSION = "branch-heads/12.2"


def local_path(path="."):
    """Return path relative to this file."""
    return abspath(pathjoin(ROOT_DIR, path))


@lru_cache(maxsize=None)
def is_win():
    return platform.startswith("win")


@lru_cache(maxsize=None)
def is_linux():
    return platform == "linux"


@lru_cache(maxsize=None)
def is_mac():
    return platform == "darwin"


class UnknownArchError(RuntimeError):
    def __init__(self, arch):
        super().__init__(f"Unknown arch {arch!r}")


@lru_cache(maxsize=None)
def get_v8_target_cpu():
    m = machine().lower()
    if m in ("arm64", "aarch64"):
        return "arm64"
    if m == "arm":
        return "arm"
    if (not m) or (match("(x|i[3-6]?)86$", m) is not None):
        return "ia32"
    if m in ("x86_64", "amd64"):
        return "x64"
    if m == "s390x":
        return "s390x"
    if m == "ppc64":
        return "ppc64"

    raise UnknownArchError(m)


@lru_cache(maxsize=None)
def get_dll_filename():
    if is_mac():
        return "libmini_racer.dylib"

    if is_win():
        return "mini_racer.dll"

    return "libmini_racer.so"


@lru_cache(maxsize=None)
def get_data_files_list():
    """List the files which v8 builds and then needs at runtime."""

    return (
        # V8 i18n data:
        "icudtl.dat",
        # V8 fast-startup snapshot; a dump of the heap after loading built-in JS
        # modules:
        "snapshot_blob.bin",
        # And obviously, the V8 build itself:
        get_dll_filename(),
    )


@lru_cache(maxsize=None)
def is_musl():
    # Alpine uses musl for libc, instead of glibc. This breaks many assumptions in the
    # V8 build, so we have to reconfigure various things when running on musl libc.
    # Determining if we're on musl (or Alpine) is surprisingly complicated; the best
    # way seems to be to check the dynamic linker ependencies of the current Python
    # executable for musl! packaging.tags.platform_tags (which is used by pip et al)
    # does this for us:
    return any("musllinux" in t for t in platform_tags())


@lru_cache(maxsize=None)
def get_platform_tag():
    """Return a pip platform tag indicating compatibility of the mini_racer binary.

    See https://packaging.python.org/en/latest/specifications/platform-compatibility-tags/.
    """

    if is_mac():
        # pip seems finicky about platform tags with larger macos versions, so just
        # tell arm64 is 11.0 and everything is is 10.9:
        if get_v8_target_cpu() == "arm64":
            return "macosx_11_0_arm64"

        return "macosx_10_9_x86_64"

    # return the first, meaning the most-specific, platform tag:
    return next(platform_tags())


@lru_cache(maxsize=None)
def get_workspace_path():
    return local_path(pathjoin("..", "v8_workspace"))


@lru_cache(maxsize=None)
def get_depot_tools_path():
    return pathjoin(get_workspace_path(), "depot_tools")


@lru_cache(maxsize=None)
def get_v8_path():
    return pathjoin(get_workspace_path(), "v8")


def unlink_if_exists(f):
    if exists(f):
        unlink(f)


def run(*args, cwd, depot_tools_first=True):
    LOGGER.debug("Calling: '%s' from working directory %s", shlexjoin(args), cwd)
    env = environ.copy()

    if depot_tools_first:
        env["PATH"] = pathsep.join([get_depot_tools_path(), environ["PATH"]])
    else:
        env["PATH"] = pathsep.join([environ["PATH"], get_depot_tools_path()])

    env["DEPOT_TOOLS_WIN_TOOLCHAIN"] = "0"
    # vpython is V8's Python environment manager; it downloads Python binaries
    # dynamically. This doesn't work on Alpine (because it downloads a glibc binary,
    # but needs a musl binary), so let's just disable it on all environments:
    env["VPYTHON_BYPASS"] = "manually managed python not supported by chrome operations"
    # Goma is a remote build system which we aren't using. depot_tools/autoninja.py
    # tries to run the goma client, which is checked into depot_tools as a glibc binary.
    # This fails on musl (on Alpine), so let's just disable the thing:
    env["GOMA_DISABLED"] = "1"

    return check_call(args, env=env, cwd=cwd)


def ensure_depot_tools():
    if isdir(get_depot_tools_path()):
        LOGGER.debug("Using already cloned depot tools")
        return

    LOGGER.debug("Cloning depot tools")
    makedirs(f"{get_workspace_path()}", exist_ok=True)
    run(
        "git",
        "clone",
        "https://chromium.googlesource.com/chromium/tools/depot_tools.git",
        cwd=get_workspace_path(),
    )

    # depot_tools will auto-update when we run various commands. This creates extra
    # dependencies, e.g., on goma (which has trouble running on Alpine due to musl).
    # We just created a fresh single-use depot_tools checkout. There is no reason to
    # update it, so let's just disable that functionality:
    open(pathjoin(get_depot_tools_path(), ".disable_auto_update"), "w").close()

    if is_win():
        # Create git.bit and maybe other shortcuts used by the Windows V8 build tools:
        run(
            pathjoin(get_depot_tools_path(), "bootstrap", "win_tools.bat"),
            cwd=get_depot_tools_path(),
        )


def ensure_v8_src(revision):
    """Ensure that v8 src are present and up-to-date."""

    # Note that we run fetch.py and gclient.py directly instead of the bash+batch
    # wrappers so that we use the correct virtualized Python.
    if not isfile(pathjoin(get_workspace_path(), ".gclient")):
        makedirs(get_workspace_path(), exist_ok=True)
        run(
            executable,
            pathjoin(get_depot_tools_path(), "fetch.py"),
            "--nohooks",
            "v8",
            cwd=get_workspace_path(),
        )

    run(
        executable,
        pathjoin(get_depot_tools_path(), "gclient.py"),
        "sync",
        "--revision",
        f"v8@{revision}",
        cwd=get_workspace_path(),
    )

    ensure_symlink(
        local_path(pathjoin("..", "src", "v8_py_frontend")),
        pathjoin(get_v8_path(), "custom_deps", "mini_racer"),
    )


def apply_patch(path, patch_filename):
    applied_patches_filename = local_path(".applied_patches")

    if not exists(applied_patches_filename):
        open(applied_patches_filename, "w").close()

    with open(applied_patches_filename, "r+") as f:
        applied_patches = set(f.read().splitlines())

        if patch_filename in applied_patches:
            return

        run(
            "patch",
            "-p0",
            "-i",
            patch_filename,
            cwd=path,
        )

        f.write(patch_filename + "\n")


def run_build(build_dir):
    """Run the actual v8 build."""

    # As explained in the design principles in ARCHITECTURE.md, we want to reduce the
    # surface area of the V8 build system which PyMiniRacer depends upon. To accomodate
    # that goal, we run with as few non-default build options as possible.

    # The standard developer guide for V8 suggests we use the v8/tools/dev/v8gen.py
    # tool to both generate args.gn, and run gn to generate Ninja build files.

    # Unfortunately, v8/tools/dev/v8gen.py is unhappy about being run on non-amd64
    # architecture (it seems to think we're always cross-compiling from amd64). Thus
    # we reproduce what it does, which for our simple case devolves to just generating
    # args.gn with minimal arguments, and running "gn gen ...".

    opts = {
        # These following settings are based those found for the "x64.release"
        # configuration. This can be verified by running:
        # tools/mb/mb.py lookup -b x64.release -m developer_default
        # ... from within the v8 directory.
        "dcheck_always_on": "false",
        "is_debug": "false",
        "target_cpu": f'"{get_v8_target_cpu()}"',
        "v8_target_cpu": f'"{get_v8_target_cpu()}"',
        # We sneak our C++ frontend into V8 as a symlinked "custom_dep" so
        # that we can reuse the V8 build system to make our dynamic link
        # library:
        "v8_custom_deps": '"//custom_deps/mini_racer"',
    }

    if (is_linux() and get_v8_target_cpu() == "arm64") or is_musl():
        # The V8 build process includes its own clang binary, but not for aarch64 on
        # Linux glibc, and not for Alpine (musl) at all.
        # Per tools/dev/gm.py, use the the system clang instead:
        opts["clang_base_path"] = '"/usr"'

        opts["clang_use_chrome_plugins"] = "false"
        # Because we use a different clang, more warnings pop up. Ignore them:
        opts["treat_warnings_as_errors"] = "false"

        # V8 currently uses a clang flag -split-threshold-for-reg-with-hint=0 which
        # doen't exist on Alpine's mainline llvm yet. Disable it:
        if is_musl():
            apply_patch(
                get_v8_path(),
                local_path("split-threshold-for-reg-with-hint.patch"),
            )

    if is_musl() and get_v8_target_cpu() == "arm64":
        # The V8 build unhelpfully sets the clang flag --target=aarch64-linux-gnu
        # on musl. The --target flag is useful when we're cross-compiling (which we're
        # not) and we aren't on aarch64-linux-gnu, we're actually on what clang calls
        # aarch64-alpine-linux-musl.
        # This patch just disables the spurious cflags and ldflags:
        apply_patch(
            get_v8_path(),
            local_path("no-aarch64-linux-gnu-target.patch"),
        )

    if is_musl():
        # On various OSes, the V8 build process brings in a whole copy of the sysroot
        # (/usr/include, /usr/lib, etc). Unfortunately on Alpine it tries to use a
        # Debian sysroot, which doesn't work. Disable it:
        opts["use_sysroot"] = "false"

        # V8 includes its own libc++ whose headers don't seem to work on Alpine:
        opts["use_custom_libcxx"] = "false"

    # We optionally use SCCACHE to speed up builds (or restart them on failure):
    sccache_path = environ.get("SCCACHE_PATH")
    if sccache_path is not None:
        opts["cc_wrapper"] = f'"{sccache_path}"'

    makedirs(build_dir, exist_ok=True)

    with open(pathjoin(build_dir, "args.gn"), "w") as f:
        f.write("# This file is auto-generated by v8_build.py")
        f.write("\n".join(f"{n}={v}" for n, v in opts.items()))
        f.write("\n")

    # Now generate Ninja build files:
    if is_musl():
        # depot_tools doesn't include a musl-compatible GN, so use the system one:
        gn_bin = ("/usr/bin/gn",)
    else:
        gn_bin = (
            executable,
            pathjoin(get_depot_tools_path(), "gn.py"),
        )

    run(
        *gn_bin,
        "gen",
        build_dir,
        "--check",
        cwd=get_v8_path(),
    )

    # Finally, actually do the build:
    if is_musl():
        # depot_tools doesn't include a musl-compatible ninja, so use the system one:
        ninja_bin = ("/usr/bin/ninja",)
    else:
        ninja_bin = (
            executable,
            pathjoin(get_depot_tools_path(), "ninja.py"),
        )

    run(
        *ninja_bin,
        # "-vv",  # this is so spammy GitHub Actions struggles to show all the output
        "-C",
        build_dir,
        pathjoin("custom_deps", "mini_racer"),
        cwd=get_v8_path(),
    )


def ensure_symlink(target, link_name):
    LOGGER.debug("Creating symlink to %s on %s", target, link_name)
    try:
        symlink(target, link_name)
    except OSError as e:
        if e.errno == EEXIST:
            remove(link_name)
            symlink(target, link_name)
        else:
            raise


def build_v8(
    out_path,
    *,
    revision=None,
    fetch_only=False,
    skip_fetch=False,
):
    revision = revision or V8_VERSION

    ensure_depot_tools()

    if not skip_fetch:
        ensure_v8_src(revision)

    if fetch_only:
        return

    build_dir = pathjoin(get_v8_path(), "out.gn", "build")

    run_build(build_dir)

    # Fish out the build artifacts:
    makedirs(out_path, exist_ok=True)

    # Create a map of actual files to in-package filenames, for Hatch to use
    # when building the wheel:
    artifacts = {}

    for f in get_data_files_list():
        src = pathjoin(build_dir, f)
        dst = pathjoin(out_path, f)

        LOGGER.debug("Copying build artifact %s to %s", src, dst)
        unlink_if_exists(dst)
        copyfile(src, dst)

        artifacts[dst] = pathjoin("py_mini_racer", f)

    LOGGER.debug("Build complete!")

    return artifacts


def clean_v8(out_path):
    for f in get_data_files_list():
        unlink_if_exists(pathjoin(out_path, f))

    rmtree(get_workspace_path(), ignore_errors=True)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
        "--out-path",
        default=pathjoin("src", "py_mini_racer"),
        help="Build destination directory",
    )
    parser.add_argument("--v8-revision", default=V8_VERSION)
    parser.add_argument("--fetch-only", action="store_true", help="Only fetch V8")
    parser.add_argument("--skip-fetch", action="store_true", help="Do not fetch V8")
    args = parser.parse_args()
    build_v8(
        out_path=args.out_path,
        revision=args.v8_revision,
        fetch_only=args.fetch_only,
        skip_fetch=args.skip_fetch,
    )



"""Hooks for mkdocs.

See https://www.mkdocs.org/user-guide/configuration/#hooks.
"""


def on_page_markdown(markdown, **kwargs):
    del kwargs

    # The markdown files in the root of the repo refer to each other by name so that
    # links between them work on the GitHub code view.
    # However, when we bring these pages into the mkdocs site, we lower-case the files
    # so the URLs look better. To make the cross-links between the pages work, do light
    # surgery on them:
    return markdown.replace("ARCHITECTURE.md", "architecture.md").replace(
        "CONTRIBUTING.md", "contributing.md"
    )



from __future__ import annotations

import ctypes
import json
import sys
from contextlib import ExitStack, contextmanager
from datetime import datetime, timezone
from importlib import resources
from json import JSONEncoder
from os.path import exists
from os.path import join as pathjoin
from sys import platform, version_info
from threading import Condition, Lock
from typing import Any, ClassVar, Iterable, Iterator, Union

Numeric = Union[int, float]


class MiniRacerBaseException(Exception):  # noqa: N818
    """Base MiniRacer exception."""


class LibNotFoundError(MiniRacerBaseException):
    """MiniRacer-wrapped V8 build not found."""

    def __init__(self, path):
        super().__init__(f"Native library or dependency not available at {path}")


class LibAlreadyInitializedError(MiniRacerBaseException):
    """MiniRacer-wrapped V8 build not found."""

    def __init__(self):
        super().__init__(
            "MiniRacer was already initialized before the call to init_mini_racer"
        )


class JSParseException(MiniRacerBaseException):
    """JavaScript could not be parsed."""


class JSEvalException(MiniRacerBaseException):
    """JavaScript could not be executed."""


class JSOOMException(JSEvalException):
    """JavaScript execution ran out of memory."""


class JSTerminatedException(JSEvalException):
    """JavaScript execution terminated."""


class JSTimeoutException(JSEvalException):
    """JavaScript execution timed out."""

    def __init__(self):
        super().__init__("JavaScript was terminated by timeout")


class JSConversionException(MiniRacerBaseException):
    """JavaScript type could not be converted."""


class WrongReturnTypeException(MiniRacerBaseException):
    """Invalid type returned by the JavaScript runtime."""

    def __init__(self, typ):
        super().__init__(f"Unexpected return value type {typ}")


class JSObject:
    """JavaScript object."""

    def __init__(self, obj):
        self.id = obj

    def __hash__(self):
        return self.id


class JSFunction:
    """JavaScript function."""


class JSSymbol:
    """JavaScript symbol."""


def _get_lib_filename(name):
    """Return the path of the library called `name` on the current system."""
    if platform == "darwin":
        prefix, ext = "lib", ".dylib"
    elif platform == "win32":
        prefix, ext = "", ".dll"
    else:
        prefix, ext = "lib", ".so"

    return prefix + name + ext


class _MiniRacerValueUnion(ctypes.Union):
    _fields_: ClassVar[tuple[str, object]] = [
        ("ptr_val", ctypes.c_void_p),
        ("bytes_val", ctypes.POINTER(ctypes.c_char)),
        ("int_val", ctypes.c_int64),
        ("double_val", ctypes.c_double),
    ]


class _MiniRacerValueStruct(ctypes.Structure):
    _fields_: ClassVar[tuple[str, object]] = [
        ("value", _MiniRacerValueUnion),
        ("len", ctypes.c_size_t),
        ("type", ctypes.c_uint8),
    ]
    _pack_ = 1


class ArrayBufferByte(ctypes.Structure):
    # Cannot use c_ubyte directly because it uses <B
    # as an internal type but we need B for memoryview.
    _fields_: ClassVar[tuple[str, object]] = [
        ("b", ctypes.c_ubyte),
    ]
    _pack_ = 1


_MR_CALLBACK = ctypes.CFUNCTYPE(
    None, ctypes.py_object, ctypes.POINTER(_MiniRacerValueStruct)
)


class _MiniRacerValHolder:
    """An object which holds open a Python reference to a _MiniRacerValueStruct owned by
    a C++ MiniRacer context."""

    def __init__(self, ctx, ptr):
        self.ctx = ctx
        self.ptr = ptr

    def __del__(self):
        self.ctx._free(self.ptr)  # noqa: SLF001


def _build_dll_handle(dll_path) -> ctypes.CDLL:
    handle = ctypes.CDLL(dll_path)

    handle.mr_init_v8.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p]

    handle.mr_init_context.argtypes = []
    handle.mr_init_context.restype = ctypes.c_void_p

    handle.mr_eval.argtypes = [
        ctypes.c_void_p,
        ctypes.c_char_p,
        ctypes.c_uint64,
        _MR_CALLBACK,
        ctypes.py_object,
    ]
    handle.mr_eval.restype = ctypes.c_void_p

    handle.mr_free_value.argtypes = [ctypes.c_void_p, ctypes.c_void_p]

    handle.mr_free_context.argtypes = [ctypes.c_void_p]

    handle.mr_free_task_handle.argtypes = [ctypes.c_void_p]

    handle.mr_heap_stats.argtypes = [
        ctypes.c_void_p,
        _MR_CALLBACK,
        ctypes.py_object,
    ]
    handle.mr_heap_stats.restype = ctypes.c_void_p

    handle.mr_low_memory_notification.argtypes = [ctypes.c_void_p]

    handle.mr_heap_snapshot.argtypes = [
        ctypes.c_void_p,
        _MR_CALLBACK,
        ctypes.py_object,
    ]
    handle.mr_heap_snapshot.restype = ctypes.c_void_p

    handle.mr_set_hard_memory_limit.argtypes = [ctypes.c_void_p, ctypes.c_size_t]

    handle.mr_set_soft_memory_limit.argtypes = [ctypes.c_void_p, ctypes.c_size_t]
    handle.mr_set_soft_memory_limit.restype = None

    handle.mr_hard_memory_limit_reached.argtypes = [ctypes.c_void_p]
    handle.mr_hard_memory_limit_reached.restype = ctypes.c_bool

    handle.mr_soft_memory_limit_reached.argtypes = [ctypes.c_void_p]
    handle.mr_soft_memory_limit_reached.restype = ctypes.c_bool

    handle.mr_v8_version.restype = ctypes.c_char_p

    handle.mr_full_eval_call_count.argtypes = [ctypes.c_void_p]
    handle.mr_full_eval_call_count.restype = ctypes.c_uint64
    handle.mr_function_eval_call_count.argtypes = [ctypes.c_void_p]
    handle.mr_function_eval_call_count.restype = ctypes.c_uint64

    return handle


# V8 internationalization data:
_ICU_DATA_FILENAME = "icudtl.dat"

# V8 fast-startup snapshot; a dump of the heap after loading built-in JS
# modules:
_SNAPSHOT_FILENAME = "snapshot_blob.bin"

DEFAULT_V8_FLAGS = ("--single-threaded",)


def _open_resource_file(filename, exit_stack):
    if version_info >= (3, 9):
        # resources.as_file was added in Python 3.9
        resource_path = resources.files("py_mini_racer") / filename

        context_manager = resources.as_file(resource_path)
    else:
        # now-deprecated API for Pythons older than 3.9
        context_manager = resources.path("py_mini_racer", filename)

    return str(exit_stack.enter_context(context_manager))


def _check_path(path):
    if path is None or not exists(path):
        raise LibNotFoundError(path)


@contextmanager
def _open_dll(flags: Iterable[str]) -> Iterator[ctypes.CDLL]:
    dll_filename = _get_lib_filename("mini_racer")

    with ExitStack() as exit_stack:
        # Find the dll and its external dependency files:
        meipass = getattr(sys, "_MEIPASS", None)
        if meipass is not None:
            # We are running under PyInstaller
            dll_path = pathjoin(meipass, dll_filename)
            icu_data_path = pathjoin(meipass, _ICU_DATA_FILENAME)
            snapshot_path = pathjoin(meipass, _SNAPSHOT_FILENAME)
        else:
            dll_path = _open_resource_file(dll_filename, exit_stack)
            icu_data_path = _open_resource_file(_ICU_DATA_FILENAME, exit_stack)
            snapshot_path = _open_resource_file(_SNAPSHOT_FILENAME, exit_stack)

        _check_path(dll_path)
        _check_path(icu_data_path)
        _check_path(snapshot_path)

        handle = _build_dll_handle(dll_path)

        handle.mr_init_v8(
            " ".join(flags).encode("utf-8"),
            icu_data_path.encode("utf-8"),
            snapshot_path.encode("utf-8"),
        )

        yield handle


_init_lock = Lock()
_dll_handle_context_manager = None
_dll_handle = None


def init_mini_racer(
    *, flags: Iterable[str] = DEFAULT_V8_FLAGS, ignore_duplicate_init=False
) -> ctypes.CDLL:
    """Initialize py_mini_racer (and V8).

    This function can optionally be used to set V8 flags. This function can be called
    at most once, before any instances of MiniRacer are initialized. Instances of
    MiniRacer will automatically call this function to initialize MiniRacer and V8.
    """

    global _dll_handle_context_manager  # noqa: PLW0603
    global _dll_handle  # noqa: PLW0603

    with _init_lock:
        if _dll_handle is None:
            _dll_handle_context_manager = _open_dll(flags)
            _dll_handle = _dll_handle_context_manager.__enter__()
            # Note: we never call _dll_handle_context_manager.__exit__() because it's
            # designed as a singleton. But we could if we wanted to!
        elif not ignore_duplicate_init:
            raise LibAlreadyInitializedError

        return _dll_handle


# Emulating asyncio.Future without asyncio.
# (Or, concurrent.futures.Future without an executor)
class _SyncFuture:
    def __init__(self):
        self._cv = Condition()
        self._complete = False
        self._res = None
        self._exc = None

    def get(self, *, timeout: Numeric | None = None):
        with self._cv:
            while not self._complete:
                if not self._cv.wait(timeout=timeout):
                    raise JSTimeoutException

            if self._exc:
                raise self._exc
            return self._res

    def set_result(self, res):
        with self._cv:
            self._res = res
            self._complete = True
            self._cv.notify()

    def set_exception(self, exc):
        with self._cv:
            self._exc = exc
            self._complete = True
            self._cv.notify()


class MiniRacer:
    """
    MiniRacer evaluates JavaScript code using a V8 isolate.

    Attributes:
        json_impl: JSON module used by helper methods default is
            [json](https://docs.python.org/3/library/json.html)
    """

    json_impl: ClassVar[object] = json

    def __init__(self):
        self._dll: ctypes.CDLL = init_mini_racer(ignore_duplicate_init=True)
        self.ctx = self._dll.mr_init_context()
        self._active_callbacks = {}

        # define an all-purpose callback:
        @_MR_CALLBACK
        def mr_callback(future, result):
            future: _SyncFuture = self._active_callbacks.pop(future)
            try:
                value = self._mr_val_to_python(result)
                future.set_result(value)
            except Exception as exc:  # noqa: BLE001
                future.set_exception(exc)

        self._mr_callback = mr_callback

    @property
    def v8_version(self) -> str:
        """Return the V8 version string."""
        return str(self._dll.mr_v8_version())

    def eval(  # noqa: A003
        self,
        code: str,
        timeout: Numeric | None = None,
        timeout_sec: Numeric | None = None,
        max_memory: int | None = None,
    ) -> Any:
        """Evaluate JavaScript code in the V8 isolate.

        Side effects from the JavaScript evaluation is persisted inside a context
        (meaning variables set are kept for the next evaluations).

        The JavaScript value returned by the last expression in `code` is converted to
        a Python value and returned by this method. Only primitive types are supported
        (numbers, strings, buffers...). Use the
        [py_mini_racer.py_mini_racer.MiniRacer.execute][] method to return more complex
        types such as arrays or objects.

        The evaluation can be interrupted by an exception for several reasons: a limit
        was reached, the code could not be parsed, a returned value could not be
        converted to a Python value.

        Args:
            code: JavaScript code
            timeout: number of milliseconds after which the execution is interrupted.
                This is deprecated; use timeout_sec instead.
            timeout_sec: number of seconds after which the execution is interrupted
            max_memory: hard memory limit, in bytes, after which the execution is
                interrupted.
        """

        if max_memory is not None:
            self.set_hard_memory_limit(max_memory)

        if timeout:
            # PyMiniRacer unfortunately uses milliseconds while Python and
            # Système international d'unités use seconds.
            timeout_sec = timeout / 1000

        if isinstance(code, str):
            code = code.encode("utf-8")

        def task(callback, future):
            return self._dll.mr_eval(
                self.ctx,
                code,
                len(code),
                callback,
                future,
            )

        with self._run_task(task) as future:
            return future.get(timeout=timeout_sec)

    def execute(
        self,
        expr: str,
        timeout: Numeric | None = None,
        timeout_sec: Numeric | None = None,
        max_memory: int | None = None,
    ) -> dict:
        """Helper to evaluate a JavaScript expression and return composite types.

        Returned value is serialized to JSON inside the V8 isolate and deserialized
        using `json_impl`.

        Args:
            expr: JavaScript expression
            timeout: number of milliseconds after which the execution is interrupted.
                This is deprecated; use timeout_sec instead.
            timeout_sec: number of seconds after which the execution is interrupted
            max_memory: hard memory limit, in bytes, after which the execution is
                interrupted.
        """

        if timeout:
            # PyMiniRacer unfortunately uses milliseconds while Python and
            # Système international d'unités use seconds.
            timeout_sec = timeout / 1000

        wrapped_expr = "JSON.stringify((function(){return (%s)})())" % expr
        ret = self.eval(wrapped_expr, timeout_sec=timeout_sec, max_memory=max_memory)
        if not isinstance(ret, str):
            raise WrongReturnTypeException(type(ret))
        return self.json_impl.loads(ret)

    def call(
        self,
        expr: str,
        *args,
        encoder: JSONEncoder | None = None,
        timeout: Numeric | None = None,
        timeout_sec: Numeric | None = None,
        max_memory: int | None = None,
    ) -> dict:
        """Helper to call a JavaScript function and return compositve types.

        The `expr` argument refers to a JavaScript function in the current V8
        isolate context. Further positional arguments are serialized using the JSON
        implementation `json_impl` and passed to the JavaScript function as arguments.

        Returned value is serialized to JSON inside the V8 isolate and deserialized
        using `json_impl`.

        Args:
            expr: JavaScript expression referring to a function
            encoder: Custom JSON encoder
            timeout: number of milliseconds after which the execution is
                interrupted.
            timeout_sec: number of seconds after which the execution is interrupted
            max_memory: hard memory limit, in bytes, after which the execution is
                interrupted
        """

        if timeout:
            # PyMiniRacer unfortunately uses milliseconds while Python and
            # Système international d'unités use seconds.
            timeout_sec = timeout / 1000

        json_args = self.json_impl.dumps(args, separators=(",", ":"), cls=encoder)
        js = f"{expr}.apply(this, {json_args})"
        return self.execute(js, timeout_sec=timeout_sec, max_memory=max_memory)

    def set_hard_memory_limit(self, limit: int) -> None:
        """Set a hard memory limit on this V8 isolate.

        JavaScript execution will be terminated when this limit is reached.

        :param int limit: memory limit in bytes or 0 to reset the limit
        """
        self._dll.mr_set_hard_memory_limit(self.ctx, limit)

    def set_soft_memory_limit(self, limit: int) -> None:
        """Set a soft memory limit on this V8 isolate.

        The Garbage Collection will use a more aggressive strategy when
        the soft limit is reached but the execution will not be stopped.

        :param int limit: memory limit in bytes or 0 to reset the limit
        """
        self._dll.mr_set_soft_memory_limit(self.ctx, limit)

    def was_hard_memory_limit_reached(self) -> bool:
        """Return true if the hard memory limit was reached on the V8 isolate."""
        return self._dll.mr_hard_memory_limit_reached(self.ctx)

    def was_soft_memory_limit_reached(self) -> bool:
        """Return true if the soft memory limit was reached on the V8 isolate."""
        return self._dll.mr_soft_memory_limit_reached(self.ctx)

    def low_memory_notification(self) -> None:
        """Ask the V8 isolate to collect memory more aggressively."""
        self._dll.mr_low_memory_notification(self.ctx)

    def heap_stats(self) -> dict:
        """Return the V8 isolate heap statistics."""

        def task(callback, future):
            return self._dll.mr_heap_stats(
                self.ctx,
                callback,
                future,
            )

        with self._run_task(task) as future:
            res = future.get()

        return self.json_impl.loads(res)

    def heap_snapshot(self) -> dict:
        """Return a snapshot of the V8 isolate heap."""

        def task(callback, future):
            return self._dll.mr_heap_snapshot(
                self.ctx,
                callback,
                future,
            )

        with self._run_task(task) as future:
            return future.get()

    def _function_eval_call_count(self) -> int:
        """Return the number of shortcut function-like evaluations done in this context.

        This is exposed for testing only."""
        return self._dll.mr_function_eval_call_count(self.ctx)

    def _full_eval_call_count(self) -> int:
        """Return the number of full script evaluations done in this context.

        This is exposed for testing only."""
        return self._dll.mr_full_eval_call_count(self.ctx)

    def _mr_val_to_python(self, ptr) -> Any:
        free_value = True
        try:
            typ = ptr.contents.type
            val = ptr.contents.value
            length = ptr.contents.len

            error_info = _ERRORS.get(ptr.contents.type)
            if error_info:
                klass, generic_msg = error_info

                msg = val.bytes_val[0:length].decode("utf-8")
                msg = msg or generic_msg

                raise klass(msg)

            if typ == MiniRacerTypes.null:
                return None
            if typ == MiniRacerTypes.bool:
                return val.int_val == 1
            if typ == MiniRacerTypes.integer:
                return val.int_val
            if typ == MiniRacerTypes.double:
                return val.double_val
            if typ == MiniRacerTypes.str_utf8:
                return val.bytes_val[0:length].decode("utf-8")
            if typ == MiniRacerTypes.function:
                return JSFunction()
            if typ == MiniRacerTypes.date:
                timestamp = val.double_val
                # JS timestamp are milliseconds, in python we are in seconds
                return datetime.fromtimestamp(timestamp / 1000.0, timezone.utc)
            if typ == MiniRacerTypes.symbol:
                return JSSymbol()
            if typ in (MiniRacerTypes.shared_array_buffer, MiniRacerTypes.array_buffer):
                cdata = (ArrayBufferByte * length).from_address(val.ptr_val)
                free_value = False
                # Keep a reference to prevent garbage collection of the backing store:
                cdata._origin = _MiniRacerValHolder(self, ptr)  # noqa: SLF001
                result = memoryview(cdata)
                # Avoids "NotImplementedError: memoryview: unsupported format T{<B:b:}"
                # in Python 3.12:
                return result.cast("B")
            if typ == MiniRacerTypes.object:
                # The C++ side puts the hash of the JS object into the value:
                return JSObject(val.int_val)

            raise JSConversionException
        finally:
            if free_value:
                self._free(ptr)

    def _free(self, res: _MiniRacerValueStruct) -> None:
        self._dll.mr_free_value(self.ctx, res)

    @contextmanager
    def _run_task(self, task):
        """Manages those tasks which generate callbacks from the MiniRacer DLL.

        Several MiniRacer functions (JS evaluation and 2 heap stats calls) are
        asynchronous. They take a function callback and callback data parameter, and
        they return a task handle.

        In this method, we create a future for each callback to get the right data to
        the right caller, and we manage the lifecycle of the task and task handle.
        """

        future = _SyncFuture()

        # Keep track of this future until we are called back.
        # This helps ensure Python doesn't garbage collect the future before the C++
        # side of things is done with it.
        self._active_callbacks[future] = future

        # Start the task:
        task_handle = task(self._mr_callback, future)
        try:
            # Let the caller handle waiting on the result:
            yield future
        finally:
            # Free the task handle.
            # Note this also cancels the task if it hasn't completed yet.
            self._dll.mr_free_task_handle(task_handle)

    def __del__(self):
        dll = getattr(self, "_dll", None)
        if dll:
            dll.mr_free_context(getattr(self, "ctx", None))


# Compatibility with versions 0.4 & 0.5
StrictMiniRacer = MiniRacer


class MiniRacerTypes:
    """MiniRacer types identifier

    Note: it needs to be coherent with mini_racer.cc.
    """

    invalid = 0
    null = 1
    bool = 2  # noqa: A003
    integer = 3
    double = 4
    str_utf8 = 5
    # deprecated:
    array = 6
    # deprecated:
    hash = 7  # noqa: A003
    date = 8
    symbol = 9
    object = 10  # noqa: A003

    function = 100
    shared_array_buffer = 101
    array_buffer = 102

    execute_exception = 200
    parse_exception = 201
    oom_exception = 202
    timeout_exception = 203
    terminated_exception = 204


_ERRORS = {
    MiniRacerTypes.parse_exception: (
        JSParseException,
        "Unknown JavaScript error during parse",
    ),
    MiniRacerTypes.execute_exception: (
        JSEvalException,
        "Uknown JavaScript error during execution",
    ),
    MiniRacerTypes.oom_exception: (JSOOMException, "JavaScript memory limit reached"),
    MiniRacerTypes.terminated_exception: (
        JSTerminatedException,
        "JavaScript was terminated",
    ),
}



import pytest
from py_mini_racer import LibAlreadyInitializedError, init_mini_racer


def test_init():
    init_mini_racer(ignore_duplicate_init=True)

    with pytest.raises(LibAlreadyInitializedError):
        init_mini_racer()

    init_mini_racer(ignore_duplicate_init=True)


# Unfortunately while init_mini_racer allows changing V8 flags, it's hard to test
# automatically because only the first use of V8 can set flags. We'd need to
# restart Python between tests.
# Here's a manual test:
# def test_init_flags():
#     from py_mini_racer import DEFAULT_V8_FLAGS, MiniRacer, init_mini_racer
#     init_mini_racer(flags=(*DEFAULT_V8_FLAGS, '--no-use-strict'))
#     mr = MiniRacer()
#     # this would normally fail in strict JS mode because foo is not declared:
#     mr.eval('foo = 4')



""" Basic JS types tests """

from datetime import datetime, timezone
from json import dumps
from time import time

import pytest
from py_mini_racer import JSEvalException, JSFunction, JSObject, MiniRacer


class Validator:
    def __init__(self):
        self.mr = MiniRacer()

    def validate(self, py_val):
        testee = py_val
        js_str = dumps(py_val)

        parsed = self.mr.execute(js_str)
        assert testee == parsed


def test_str():
    v = Validator()
    v.validate("'a string'")
    v.validate("'a ' + 'string'")


def test_unicode():
    ustr = "\N{GREEK CAPITAL LETTER DELTA}"
    mr = MiniRacer()
    res = mr.eval("'" + ustr + "'")
    assert ustr == res


def test_numbers():
    v = Validator()
    v.validate(1)
    v.validate(1.0)
    v.validate(2**16)
    v.validate(2**31 - 1)
    v.validate(2**31)
    v.validate(2**33)


def test_arrays():
    v = Validator()
    v.validate([1])
    v.validate([])
    v.validate([1, 2, 3])
    # Nested
    v.validate([1, 2, ["a", 1]])


def test_none():
    v = Validator()
    v.validate(None)


def test_hash():
    v = Validator()
    v.validate({})
    v.validate("{}")
    v.validate({"a": 1})
    v.validate({" ": {"z": "www"}})


def test_complex():
    v = Validator()
    v.validate(
        {
            "1": [
                1,
                2,
                "qwe",
                {"z": [4, 5, 6, {"eqewr": 1, "zxczxc": "qweqwe", "z": {"1": 2}}]},
            ],
            "qwe": 1,
        }
    )


def test_function():
    mr = MiniRacer()
    res = mr.eval("var a = function(){}; a")
    assert isinstance(res, JSFunction)


def test_object():
    mr = MiniRacer()
    res = mr.eval("var a = {}; a")
    assert isinstance(res, JSObject)


def test_timestamp():
    val = int(time())
    mr = MiniRacer()
    res = mr.eval("var a = new Date(%d); a" % (val * 1000))
    assert res == datetime.fromtimestamp(val, timezone.utc)


def test_date():
    mr = MiniRacer()
    res = mr.eval("var a = new Date(Date.UTC(2014, 0, 2, 3, 4, 5)); a")
    assert res == datetime(2014, 1, 2, 3, 4, 5, tzinfo=timezone.utc)


def test_exception():
    js_source = """
    var f = function(arg) {
        throw 'error: '+arg
        return nil
    }"""

    mr = MiniRacer()
    mr.eval(js_source)

    with pytest.raises(JSEvalException) as exc_info:
        mr.eval("f(42)")

    assert "error: 42" in exc_info.value.args[0]


def test_array_buffer():
    js_source = """
    var b = new ArrayBuffer(1024);
    var v = new Uint8Array(b);
    v[0] = 0x42;
    b
    """
    mr = MiniRacer()
    ret = mr.eval(js_source)
    assert len(ret) == 1024
    assert ret[0:1].tobytes() == b"\x42"


def test_array_buffer_view():
    js_source = """
    var b = new ArrayBuffer(1024);
    var v = new Uint8Array(b, 1, 1);
    v[0] = 0x42;
    v
    """
    mr = MiniRacer()
    ret = mr.eval(js_source)
    assert len(ret) == 1
    assert ret.tobytes() == b"\x42"


def test_shared_array_buffer():
    js_source = """
    var b = new SharedArrayBuffer(1024);
    var v = new Uint8Array(b);
    v[0] = 0x42;
    b
    """
    mr = MiniRacer()
    ret = mr.eval(js_source)
    assert len(ret) == 1024
    assert ret[0:1].tobytes() == b"\x42"
    ret[1:2] = b"\xFF"
    assert mr.eval("v[1]") == 0xFF



from py_mini_racer import py_mini_racer

DEFAULT_V8_FLAGS = py_mini_racer.DEFAULT_V8_FLAGS
JSEvalException = py_mini_racer.JSEvalException
JSFunction = py_mini_racer.JSFunction
JSOOMException = py_mini_racer.JSOOMException
JSObject = py_mini_racer.JSObject
JSParseException = py_mini_racer.JSParseException
JSSymbol = py_mini_racer.JSSymbol
JSTimeoutException = py_mini_racer.JSTimeoutException
LibAlreadyInitializedError = py_mini_racer.LibAlreadyInitializedError
LibNotFoundError = py_mini_racer.LibNotFoundError
MiniRacer = py_mini_racer.MiniRacer
StrictMiniRacer = py_mini_racer.StrictMiniRacer
init_mini_racer = py_mini_racer.init_mini_racer


__all__ = ["py_mini_racer", "MiniRacer"]



""" Test .eval() method """

from time import sleep, time

import pytest
from py_mini_racer import (
    JSEvalException,
    JSOOMException,
    JSParseException,
    JSSymbol,
    JSTimeoutException,
    MiniRacer,
)


def test_invalid():
    mr = MiniRacer()

    with pytest.raises(JSEvalException) as exc_info:
        mr.eval("invalid")

    assert (
        exc_info.value.args[0]
        == """\
<anonymous>:1: ReferenceError: invalid is not defined
invalid
^

ReferenceError: invalid is not defined
    at <anonymous>:1:1
"""
    )


def test_eval():
    mr = MiniRacer()
    assert mr.eval("42") == 42


def test_blank():
    mr = MiniRacer()
    assert mr.eval("") is None
    assert mr.eval(" ") is None
    assert mr.eval("\t") is None

    assert mr._full_eval_call_count() == 3  # noqa: SLF001


def test_global():
    mr = MiniRacer()
    mr.eval("var xabc = 22;")
    assert mr.eval("xabc") == 22


def test_fun():
    mr = MiniRacer()
    mr.eval("var x = function(y) {return y+1;}")

    assert mr.eval("x(1)") == 2
    assert mr.eval("x(10)") == 11
    assert mr.eval("x(100)") == 101


def test_multiple_ctx():
    c1 = MiniRacer()
    c2 = MiniRacer()
    c3 = MiniRacer()

    c1.eval("var x = 1")
    c2.eval("var x = 2")
    c3.eval("var x = 3")
    assert c1.eval("(x)") == 1
    assert c2.eval("(x)") == 2
    assert c3.eval("(x)") == 3


def test_exception_thrown():
    context = MiniRacer()

    js_source = "var f = function() {throw new Error('blah')};"

    context.eval(js_source)

    # Add extra junk (+'') to avoid our fast function call path which produces a
    # slightly different traceback:
    with pytest.raises(JSEvalException) as exc_info:
        context.eval("f()+''")

    assert (
        exc_info.value.args[0]
        == """\
<anonymous>:1: Error: blah
var f = function() {throw new Error('blah')};
                    ^

Error: blah
    at f (<anonymous>:1:27)
    at <anonymous>:1:1
"""
    )


def test_fast_call_exception_thrown():
    context = MiniRacer()

    js_source = "var f = function() {throw new Error('blah')};"

    context.eval(js_source)

    # This should go into our fast function call route
    with pytest.raises(JSEvalException) as exc_info:
        context.eval("f()")

    assert (
        exc_info.value.args[0]
        == """\
<anonymous>:1: Error: blah
var f = function() {throw new Error('blah')};
                    ^

Error: blah
    at f (<anonymous>:1:27)
"""
    )

    assert context._function_eval_call_count() == 1  # noqa: SLF001


def test_string_thrown():
    context = MiniRacer()

    js_source = "var f = function() {throw 'blah'};"

    context.eval(js_source)

    with pytest.raises(JSEvalException) as exc_info:
        context.eval("f()")

    # When you throw a plain string (not wrapping it in a `new Error(...)`), you get no
    # backtrace:
    assert (
        exc_info.value.args[0]
        == """\
<anonymous>:1: blah
var f = function() {throw 'blah'};
                    ^
"""
    )


def test_cannot_parse():
    context = MiniRacer()
    js_source = "var f = function("

    with pytest.raises(JSParseException) as exc_info:
        context.eval(js_source)

    assert (
        exc_info.value.args[0]
        == """\
<anonymous>:1: SyntaxError: Unexpected end of input
var f = function(
                 ^

SyntaxError: Unexpected end of input
"""
    )


def test_null_byte():
    context = MiniRacer()

    s = "\x00 my string!"

    # Try return a string including a null byte
    in_val = 'var str = "' + s + '"; str;'
    result = context.eval(in_val)
    assert result == s


def test_timeout():
    timeout = 0.1
    start_time = time()

    mr = MiniRacer()
    with pytest.raises(JSTimeoutException) as exc_info:
        mr.eval("while(1) { }", timeout_sec=timeout)

    duration = time() - start_time
    # Make sure it timed out on time, and allow a giant leeway (because aarch64
    # emulation tests are surprisingly slow!)
    assert timeout <= duration <= timeout + 5

    assert exc_info.value.args[0] == "JavaScript was terminated by timeout"


def test_timeout_ms():
    # Same as above but with the deprecated timeout millisecond argument
    timeout = 0.1
    start_time = time()

    mr = MiniRacer()
    with pytest.raises(JSTimeoutException) as exc_info:
        mr.eval("while(1) { }", timeout=int(timeout * 1000))

    duration = time() - start_time
    # Make sure it timed out on time, and allow a giant leeway (because aarch64
    # emulation tests are surprisingly slow!)
    assert timeout <= duration <= timeout + 5

    assert exc_info.value.args[0] == "JavaScript was terminated by timeout"


def test_max_memory_soft():
    mr = MiniRacer()
    mr.set_soft_memory_limit(100000000)
    mr.set_hard_memory_limit(100000000)
    with pytest.raises(JSOOMException) as exc_info:
        mr.eval(
            """let s = 1000;
            var a = new Array(s);
            a.fill(0);
            while(true) {
                s *= 1.1;
                let n = new Array(Math.floor(s));
                n.fill(0);
                a = a.concat(n);
            }""",
        )

    assert mr.was_soft_memory_limit_reached()
    assert mr.was_hard_memory_limit_reached()
    assert exc_info.value.args[0] == "JavaScript memory limit reached"


def test_max_memory_hard():
    mr = MiniRacer()
    mr.set_hard_memory_limit(100000000)
    with pytest.raises(JSOOMException) as exc_info:
        mr.eval(
            """let s = 1000;
            var a = new Array(s);
            a.fill(0);
            while(true) {
                s *= 1.1;
                let n = new Array(Math.floor(s));
                n.fill(0);
                a = a.concat(n);
            }""",
        )

    assert not mr.was_soft_memory_limit_reached()
    assert mr.was_hard_memory_limit_reached()
    assert exc_info.value.args[0] == "JavaScript memory limit reached"


def test_max_memory_hard_eval_arg():
    # Same as above but passing the argument into the eval method (which is a
    # deprecated thing to do because the parameter is really affine to the
    # MiniRacer object)
    mr = MiniRacer()
    with pytest.raises(JSOOMException) as exc_info:
        mr.eval(
            """let s = 1000;
            var a = new Array(s);
            a.fill(0);
            while(true) {
                s *= 1.1;
                let n = new Array(Math.floor(s));
                n.fill(0);
                a = a.concat(n);
            }""",
            max_memory=200000000,
        )

    assert exc_info.value.args[0] == "JavaScript memory limit reached"


def test_symbol():
    mr = MiniRacer()
    res = mr.eval("Symbol.toPrimitive")
    assert isinstance(res, JSSymbol)


def test_microtask():
    # PyMiniRacer uses V8 microtasks (things, like certain promise callbacks, which run
    # immediately after an evaluation ends).
    # By default, V8 runs any microtasks before it returns control to PyMiniRacer.
    # Let's test that they actually work.
    # PyMiniRacer does not expose the web standard `window.queueMicrotask` (because it
    # does not expose a `window` to begin with). We can, however, trigger a microtask
    # by triggering one as a side effect of a `then` on a resolved promise:
    mr = MiniRacer()
    assert not mr.eval(
        """
    let p = Promise.resolve();

    var done = false;

    p.then(() => {done = true});

    done
    """
    )
    assert mr.eval("done")


def test_async():
    mr = MiniRacer()
    assert not mr.eval(
        """
    var done = false;
    const shared = new SharedArrayBuffer(8);
    const view = new Int32Array(shared);

    const p = Atomics.waitAsync(view, 0, 0, 1000); // 1 s timeout
    p.value.then(() => { done = true; });
    done
    """
    )
    assert not mr.eval("done")
    start = time()
    # Give the 1-second wait 10 seconds to finish. (Emulated aarch64 tests are
    # surprisingly slow!)
    while time() - start < 10 and not mr.eval("done"):
        sleep(0.1)
    assert mr.eval("done")


def test_fast_call():
    mr = MiniRacer()
    mr.eval("var test = function () { return 42; }")

    # This syntax is optimized and takes another execution path in the extension
    # Note: the fast call optimization only works with functions defined using
    # "var", of "function test() { ... }", which place the definition on the global
    # "this". It doesn't work with functions defined as "let" or "const" because
    # they place the function onto block scope, where our optimization cannot find
    # them.
    assert mr.eval("test()") == 42

    # It looks like a fast call but it is not (it ends with '()' but no identifier)
    # should not fail and do a classical evaluation.
    assert mr.eval("1+test()") == 43

    assert mr._function_eval_call_count() == 1  # noqa: SLF001
    assert mr._full_eval_call_count() == 2  # noqa: SLF001



""" Test loading and executing babel.js """

from os.path import dirname
from os.path import join as pathjoin

from py_mini_racer import MiniRacer


def test_babel():
    context = MiniRacer()

    path = pathjoin(dirname(__file__), "fixtures/babel.js")
    with open(path, encoding="utf-8") as f:
        babel_source = f.read()
    source = (
        """
      var self = this;
      %s
      babel.eval = function(code) {
        return eval(babel.transform(code)["code"]);
      }
    """
        % babel_source
    )
    context.eval(source)
    assert context.eval("babel.eval(((x) => x * x)(8))") == 64



import pytest
from py_mini_racer import JSEvalException, StrictMiniRacer


def test_basic_int():
    mr = StrictMiniRacer()
    assert mr.execute("42") == 42


def test_basic_string():
    mr = StrictMiniRacer()
    assert mr.execute('"42"') == "42"


def test_basic_hash():
    mr = StrictMiniRacer()
    assert mr.execute("{}") == {}


def test_basic_array():
    mr = StrictMiniRacer()
    assert mr.execute("[1, 2, 3]") == [1, 2, 3]


def test_call():
    js_func = """var f = function(args) {
        return args.length;
    }"""

    mr = StrictMiniRacer()

    assert mr.eval(js_func) is None
    assert mr.call("f", list(range(5))) == 5


def test_message():
    mr = StrictMiniRacer()
    with pytest.raises(JSEvalException):
        mr.eval("throw new EvalError('Hello', 'someFile.js', 10);")



"""Test executing a WASM module."""

from os.path import abspath, dirname, getsize
from os.path import join as pathjoin

from py_mini_racer import MiniRacer

test_dir = dirname(abspath(__file__))


def test_add():
    fn = pathjoin(test_dir, "add.wasm")
    mr = MiniRacer()

    # 1. Allocate a buffer to hold the WASM module code
    size = getsize(fn)
    module_raw = mr.eval(
        f"""
    const moduleRaw = new SharedArrayBuffer({size});
    moduleRaw
    """
    )

    # 2. Read the WASM module code
    with open(fn, "rb") as f:
        assert f.readinto(module_raw) == size

    # 3. Instantiate the WASM module
    mr.eval(
        """
    var res = null;
    WebAssembly.instantiate(new Uint8Array(moduleRaw)).then(result => {
        res = result.instance;
    }).catch(result => { res = result.message; });
    """
    )

    # 4. Wait for WASM module instantiation
    while not mr.eval("res"):
        pass

    assert mr.eval("typeof res !== 'string'")

    # 5. Execute a WASM function
    assert mr.eval("res.exports.addTwo(1, 2)") == 3



""" Transform the input stream using babel.transform """

import os
import sys

from py_mini_racer import py_mini_racer


def babel_transform(es_string):
    """Transform the provided string using babel.transform"""

    path_to_babel = os.path.join(
        os.path.dirname(__file__), "..", "tests", "fixtures", "babel.js"
    )

    with open(path_to_babel) as f:
        babel_source = f.read()

    # Initializes PyMiniRacer
    ctx = py_mini_racer.MiniRacer()

    # Parse babel
    ctx.eval("""var self = this; %s """ % babel_source)

    # Transform stuff :)
    val = "babel.transform(`%s`)['code']" % es_string
    return ctx.eval(val)


if __name__ == "__main__":
    if len(sys.argv) != 1:
        name = sys.argv[0]
        sys.stderr.write("Usage: cat es6file.js | %s\n" % name)
        sys.stderr.write("Example: echo [1,2,3].map(n => n + 1); | %s\n" % name)
        sys.exit(-1)

    es6_data = sys.stdin.read()

    res = babel_transform(es6_data)

    sys.stdout.write(res)



""" Basic JS call functions """

from datetime import datetime, timezone
from json import JSONEncoder

from py_mini_racer import MiniRacer


def test_call_js():
    js_func = """var f = function() {
        return arguments.length;
    }"""

    mr = MiniRacer()
    mr.eval(js_func)

    assert mr.call("f") == 0
    assert mr.call("f", *list(range(5))) == 5
    assert mr.call("f", *list(range(10))) == 10
    assert mr.call("f", *list(range(20))) == 20


def test_call_custom_encoder():
    # Custom encoder for dates
    class CustomEncoder(JSONEncoder):
        def default(self, obj):
            if isinstance(obj, datetime):
                return obj.isoformat()

            return JSONEncoder.default(self, obj)

    now = datetime.now(tz=timezone.utc)
    js_func = """var f = function(args) {
        return args;
    }"""
    mr = MiniRacer()
    mr.eval(js_func)

    assert mr.call("f", now, encoder=CustomEncoder) == now.isoformat()
